use std::collections::HashMap;
use std::fs;
use directories::UserDirs;
use serde::{Deserialize, Serialize};
use crate::{DockerTags, QueryArgs};

/// registry api response struct
#[derive(Serialize, Deserialize)]
struct RegistryTagsResponse {
    name: Option<String>,
    tags: Option<Vec<String>>,
}

#[derive(Serialize, Deserialize)]
struct RegistryTokenResponse {
    token: Option<String>,
    access_token: Option<String>,
    expires_in: u32,
    issued_at: Option<String>,
}


pub struct RegistryTagsFetcher {}

impl DockerTags for RegistryTagsFetcher {
    fn get_tags(&self, args: &QueryArgs) -> Vec<String> {
        let mut namespace = "library";
        let repository = match &args.repository {
            Some(s) => {
                let vec = s.split('/')
                    .into_iter()
                    .collect::<Vec<&str>>();
                namespace = vec[0];
                vec[1]
            }
            None => { "" }
        };
        let token = get_token(namespace, repository).unwrap();

        let url = format!("https://registry-1.docker.io/v2/{namespace}/{repository}/tags/list");
        let client = reqwest::blocking::Client::new();
        let response = client.get(url)
            .header("Authorization", format!("Bearer {token}"))
            .send();
        let mut results = Vec::<String>::new();
        if let Ok(r) = response {
            let json: RegistryTagsResponse = serde_json::from_str(r.text().unwrap().as_str()).unwrap();
            results.append(&mut json.tags.unwrap());
        };
        results
    }
}

fn get_token(namespace: &str, repository: &str) -> Option<String> {
    let config = read_config();
    match config {
        Some(s) => {
            let service = "registry.docker.io";
            let scope = format!("repository:{namespace}/{repository}:pull");
            let auth = format!("https://auth.docker.io/token?service={service}&scope={scope}");
            let client = reqwest::blocking::Client::new();
            let response = client.get(auth)
                .basic_auth(s.username, Some(s.password))
                .send();
            match response {
                Ok(r) => {
                    let json: RegistryTokenResponse = serde_json::from_str(r.text().unwrap().as_str()).unwrap();
                    Some(json.token.unwrap())
                }
                Err(_) => { None }
            }
        }
        _ => { None }
    }
}

#[derive(Serialize, Deserialize)]
struct DockerConfig {
    auths: HashMap<String, DockerAuth>,
}

#[derive(Serialize, Deserialize)]
struct DockerAuth {
    auth: Option<String>,
}

use base64::{Engine as _, engine::general_purpose};

fn read_config() -> Option<DockerPassword> {
    let user_dirs = UserDirs::new().unwrap();
    let home_dir = user_dirs.home_dir();
    let config_path = home_dir.join(".docker").join("config.json");
    let config = fs::read_to_string(config_path);
    let json: DockerConfig = serde_json::from_str(config.unwrap().as_str()).unwrap();
    let auths = json.auths;
    // let auth = auths.get("https://hub.docker.com/v2/");
    let auth = auths.get("https://index.docker.io/v1/");
    match auth {
        Some(a) => {
            let auth = a.auth.clone().unwrap();

            let mut buffer = Vec::<u8>::new();
            let result = general_purpose::URL_SAFE_NO_PAD.decode_vec(auth.as_bytes(), &mut buffer);
            let decoded = match std::str::from_utf8(&buffer) {
                Ok(v) => v,
                Err(e) => panic!("Invalid UTF-8 sequence: {}", e),
            };
            let vec = decoded.split(":").collect::<Vec<&str>>();
            println!("{:?}", decoded);
            Some(DockerPassword {
                username: vec[0].parse().unwrap(),
                password: vec[1].parse().unwrap(),
            })
        }
        _ => { None }
    }
}

#[derive(Serialize, Deserialize)]
struct DockerPassword {
    username: String,
    password: String,
}